unit UConvert;
{$DEFINE GIF}
interface

uses
  Windows, Messages, SysUtils, Classes, VCL.Graphics,
  VCL.Imaging.PNGImage, VCL.Imaging.JPEG{$IFDEF GIF}, VCL.Imaging.GIFImg{$ENDIF};

function ConvertStream(const Stream: TStream; MaxSize: TPoint): HBitmap;
function GetVeryLongFileName(const FullFileName: string): string; inline;
function DestSize(const CurSize, MaxSize: TPoint): TPoint;

implementation

function GetVeryLongFileName(const FullFileName: string): string; inline;
begin
  Result := FullFileName;
  if Length(FullFileName) < MAX_PATH then Exit;
  if Pos('\\?\', FullFileName) > 0 then Exit;

  if Pos('\\', FullFileName) > 0 then
    Result := '\\?\UNC\' + Copy(FullFileName, 3, MaxInt)
  else
    Result := '\\?\' + FullFileName;
end;

function DestSize(const CurSize, MaxSize: TPoint): TPoint;
var
  w, h, cw, ch: Integer;
  xyaspect: Double;
begin
  w := CurSize.X;
  h := CurSize.Y;
  cw := MaxSize.X;
  ch := MaxSize.Y;

  // In rare case Width or Height can be calculated 1 pixel less. So check dimensions first.
  if (w = cw) and (h = ch) then
    Exit(Point(w, h));

	if (w > 0) and (h > 0) then
	begin
      xyaspect := w / h;
      if w > h then
      begin
        w := cw;
        h := Trunc(cw / xyaspect);
        if h > ch then  // woops, too big
        begin
          h := ch;
          w := Trunc(ch * xyaspect);
        end;
      end
      else
      begin
        h := ch;
        w := Trunc(ch * xyaspect);
        if w > cw then  // woops, too big
        begin
          w := cw;
          h := Trunc(cw / xyaspect);
        end;
      end;
    end;

  with Result do
  begin
    X := w;
    Y := h;
  end;
end;

function GetJpegSize(const JPEG: TStream;
  out width, height, BitDepth: Integer): boolean;
var
  n: Integer;
  b: byte;
  w: Word;
begin
  Result := false;
  try
    n := JPEG.Size - 8;
    JPEG.Position := 0;
    if n <= 0 then
      Exit;
    JPEG.Read(w, 2);
    if w <> $D8FF then
      Exit; // invalid format
    JPEG.Read(b, 1);
    while (JPEG.Position < n) and (b = $FF) do
    begin
      JPEG.Read(b, 1);
      case b of
        $C0 .. $C3:
          begin
            JPEG.Seek(3, soFromCurrent);
            JPEG.Read(w, 2);
            height := swap(w);
            JPEG.Read(w, 2);
            width := swap(w);
            JPEG.Read(b, 1);
            BitDepth := b * 8;
            Result := true; // JPEG format OK
            Exit;
          end;
        $FF:
          JPEG.Read(b, 1);
        $D0 .. $D9, $01:
          begin
            JPEG.Seek(1, soFromCurrent);
            JPEG.Read(b, 1);
          end;
      else
        begin
          JPEG.Read(w, 2);
          JPEG.Seek(swap(w) - 2, soFromCurrent);
          JPEG.Read(b, 1);
        end;
      end;
    end;
  except
    Result := false;
  end;
end;

function GetJPEGScale(const Stream: TStream; const AutoCalcWidth,
  AutoCalcHeight: Integer): TJPEGScale;
var
  w, h, b, scale, bst, dv, i: Integer;
begin
  Result := jsFullSize;
  Stream.Position := 0;
  if not GetJpegSize(Stream, w, h, b) then
    Exit;

  if (AutoCalcWidth > 0) and (AutoCalcHeight > 0) then
  begin
    bst := 0;
    for i := 3 downto 0 do
    begin
      dv := 1 shl i;
      if ((w div dv) >= AutoCalcWidth) and ((h div dv) >= AutoCalcHeight) then
      begin
        bst := i;
        break;
      end;
    end;
    scale := 1 shl bst;
  end
  else
    scale := 1;

  case scale of
    1: Result := jsFullSize;
    2: Result := jsHalf;
    4: Result := jsQuarter;
    8: Result := jsEighth;
  end;
end;

function LoadJPEGFromStream(const Stream: TStream; const AutoCalcWidth,
  AutoCalcHeight: Integer): TBitmap;
var
  JPG: TJPEGImage;
begin
  Result := nil;
  try
    JPG := TJPEGImage.Create;
    try
      JPG.Scale := GetJPEGScale(Stream, AutoCalcWidth, AutoCalcHeight);
      //MessageBox(GetActiveWindow, PChar(IntToStr(Ord(JPG.Scale))), nil, MB_OK);
      Stream.Position := 0;
      JPG.LoadFromStream(Stream);
      Result := TBitmap.Create;
      Result.Assign(JPG);
    finally
      JPG.Free;
    end;
  except
    FreeAndNil(Result);
  end;
end;

function LoadPNGFromStream(const FS: TStream;
  const SimpleRemoveTransparency: Boolean = True): TBitmap;
var
  PNG: TPngImage;
begin
  Result := nil;
  try
    PNG := TPngImage.Create;
    try
      PNG.LoadFromStream(FS);
      if SimpleRemoveTransparency then
        PNG.RemoveTransparency;

      Result := TBitmap.Create;

      if not SimpleRemoveTransparency then
      begin
        Result.PixelFormat := pf24bit;
        Result.SetSize(PNG.Width, PNG.Height);
        //PNG.Draw(Result.Canvas, Result.Canvas.ClipRect);
        PNG.Draw(Result.Canvas, TRect.Create(0, 0, PNG.Width, PNG.Height));
      end else
        Result.Assign(PNG)

    finally
      PNG.Free;
    end;
  except
    FreeAndNil(Result);
  end;
end;
{$IFDEF GIF}
function LoadGIFFromStream(const FS: TStream): TBitmap;
var
  GIF: TGIFImage;
begin
  Result := nil;
  try
    GIF := TGIFImage.Create;
    try
      GIF.LoadFromStream(FS);
      Result := TBitmap.Create;
      Result.Assign(GIF);
    finally
      GIF.Free;
    end;
  except
    FreeAndNil(Result);
  end;
end;
{$ENDIF}
function LoadBMPFromStream(const FS: TStream): TBitmap;
begin
  Result := TBitmap.Create;
  try
    Result.LoadFromStream(FS);
  except
    FreeAndNil(Result);
  end;
end;

function ConvertStream(const Stream: TStream; MaxSize: TPoint): HBitmap;
var
  BMP, TmpBMP: TBitmap;
  FileType: array[0..3] of Byte;
  NewSize: TPoint;
begin
  Result := 0;
  TmpBMP := nil;
  BMP := nil;

  try
    FillChar(FileType, SizeOf(FileType), 0);
    try
      if Stream.Read(FileType, SizeOf(FileType)) < 2 then
        Exit;

      if (FileType[0] = $FF) and (FileType[1] = $D8) then
      begin
        Stream.Position := 0;
        BMP := LoadJPEGFromStream(Stream, MaxSize.X, MaxSize.Y);
      end else
      if (FileType[0] = $89) and (FileType[1] = $50) and (FileType[2] = $4E)
         and (FileType[3] = $47) then
      begin
        Stream.Position := 0;
        BMP := LoadPNGFromStream(Stream);
      end else
      {$IFDEF GIF}
      if (FileType[0] = $47) and (FileType[1] = $49) and (FileType[2] = $46) then
      begin
        Stream.Position := 0;
        BMP := LoadGIFFromStream(Stream);
      end else
      {$ENDIF}
      if (FileType[0] = $42) and (FileType[1] = $4D) then
      begin
        Stream.Position := 0;
        BMP := LoadBMPFromStream(Stream);
      end else
        Exit;

      if not Assigned(BMP) then
        Exit;

      if (BMP.Width > MaxSize.X) or (BMP.Height > MaxSize.Y) then
      begin
        TmpBMP := TBitmap.Create;
        TmpBMP.PixelFormat := pf24bit;
        NewSize := DestSize(Point(BMP.Width, BMP.Height), MaxSize);
        TmpBMP.Width := NewSize.X;
        TmpBMP.Height := NewSize.Y;
        SetStretchBltMode(TmpBMP.Canvas.Handle, STRETCH_HALFTONE);
        SetBrushOrgEx(TmpBMP.Canvas.Handle, 0, 0, nil);
        StretchBlt(TmpBMP.Canvas.Handle, 0, 0, TmpBMP.Width, TmpBMP.Height,
          BMP.Canvas.Handle, 0, 0, BMP.Width, BMP.Height, SRCCOPY);
        BMP.Assign(TmpBMP);
      end;
      BMP.HandleType := bmDIB;
      Result := BMP.ReleaseHandle;
    finally
      BMP.Free;
      TmpBMP.Free;
    end;
  except
    //on E: Exception do
      //MessageBox(GetActiveWindow, PChar(E.Message), nil, MB_OK or MB_ICONERROR);
  end;
end;

end.
